LIBRARY     SWMM5.DLL

EXPORTS
    swmm_close                    = _swmm_close@0
    swmm_decodeDate               = _swmm_decodeDate@36
    swmm_end                      = _swmm_end@0
    swmm_getCount                 = _swmm_getCount@4
    swmm_getError                 = _swmm_getError@8
    swmm_getIndex                 = _swmm_getIndex@8
    swmm_getMassBalErr            = _swmm_getMassBalErr@12
    swmm_getName                  = _swmm_getName@16
    swmm_getSavedValue            = _swmm_getSavedValue@12
    swmm_getValue                 = _swmm_getValue@8
    swmm_getVersion               = _swmm_getVersion@0
    swmm_getWarnings              = _swmm_getWarnings@0
    swmm_open                     = _swmm_open@12
    swmm_report                   = _swmm_report@0
    swmm_run                      = _swmm_run@12
    swmm_setValue                 = _swmm_setValue@16
    swmm_start                    = _swmm_start@4
    swmm_step                     = _swmm_step@4
    swmm_stride                   = _swmm_stride@8
    swmm_writeLine                = _swmm_writeLine@4
