//-----------------------------------------------------------------------------
//   headers.h
//
//   Project: EPA SWMM5
//   Version: 5.2
//   Date:    11/01/21 (Build 5.2.0)
//   Author:  L. Rossman
//
//   Header files included in most SWMM5 modules.
//
//   DO NOT CHANGE THE ORDER OF THE #INCLUDE STATEMENTS
//-----------------------------------------------------------------------------
#include "macros.h"
#include "objects.h"
#define  EXTERN extern
#include "globals.h"
#include "funcs.h"
#include "error.h"
#include "text.h"
#include "keywords.h"
