# -*- coding: utf-8 -*-

#
#  __init__.py - SWMM toolkit package
#
#  Created:    Aug 9, 2018
#  Updated:    September 4, 2025
#
#  Author:     See AUTHORS
#

"""
A low level pythonic API for the swmm-output and swmm-solver dlls using SWIG.
"""


__author__ = "See AUTHORS"
__credits__ = "Colleen Barr"
__license__ = "CC0 1.0 Universal"

__version__ = "0.17.0"
__date__ = "December 4, 2025"

__maintainer__ = "Michael Tryby"
__email__ = "michael.tryby@gmail.com"
__status__ = "Production/Stable"
